//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
#include <master_globals.h>
#include <guiaux.h>
// GUI CONSOLE IMPLEMENTATION

BOOL ViewMenuHandler(gEvent *e,gControl *This);
BOOL ExitViewMenuHandler(gEvent *e,gControl *This);
BOOL BreakHandler(gEvent *e,gControl *This);
BOOL BreakQueryHandler(gEvent *e,gControl *This);
extern "C" void updateConsole(void);

#define ConsolePTR __mg->ConsPtr

#define Console ((STDConsole *)(__mg->ConsPtr))

const struct gSoftMenuItem ViewModeMenu[6]={ {0,"",&gDoNothingEvent},
											{0,"",&gDoNothingEvent},
											{0,"|Brk|",&BreakHandler},
											{0,"",&gDoNothingEvent},
											{0,"",&gDoNothingEvent},
											{0,"Back",&ExitViewMenuHandler}
};
const struct gSoftMenuItem NormalModeMenu[6]={ {0,"",&gDoNothingEvent},
											{0,"",&gDoNothingEvent},
											{0,"|Brk|",&BreakHandler},
											{0,"",&gDoNothingEvent},
											{0,"",&gDoNothingEvent},
											{0,"View",&ViewMenuHandler}
											};



STDConsole::STDConsole()
{
	MaxSize=16384;
	App=new gApplication();
	if(!App) {
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Font=new gFont(FONT_SYSTEM6);
	if(!Font) {
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Form=new gForm(*App,NULL,(gFormStyles) S_SINGLESOFTMENU);
	if(!Form) {
		delete App;
		delete Font;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Menu=new gSoftMenu(*Form,S_SINGLESOFTMENU);
	if(!Menu) {
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Editor=new gEditBox(*Form,0,0,Form->vieww,Form->viewh,(gStyles)(S_MULTILINE | S_VSCROLLBAR),Font);
	if(!Editor) {
		delete Menu;
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Loop=App->StartOpenLoop(Form);
	if(!Loop) {
		delete Editor;
		delete Menu;
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
		
	}
	Editor->SetReadOnly(TRUE);
	Editor->SetOverwriteMode(TRUE);
	Menu->AddItemList((gSoftMenuItem *)NormalModeMenu,6);

	
	Editor->AddHandler(GEVENT_GUI,-1,GEVGUI_QUERYCLOSE,BreakQueryHandler);

	lock=1;
	update_event=tmr_eventcreate(updateConsole,250,TRUE);
}

STDConsole::STDConsole(int fontstyle,int _MaxSize)
{
	lock=1;
	MaxSize=_MaxSize;
	App=new gApplication();
	if(!App) {
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Font=new gFont(fontstyle);
	if(!Font) {
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Form=new gForm(*App,NULL,(gFormStyles)S_SINGLESOFTMENU);
	if(!Form) {
		delete App;
		delete Font;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Menu=new gSoftMenu(*Form,S_SINGLESOFTMENU);
	if(!Menu) {
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	Editor=new gEditBox(*Form,0,0,Form->vieww,Form->viewh, (gStyles)(S_MULTILINE | S_VSCROLLBAR),Font);
	if(!Editor) {
		delete Menu;
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
	}
	
	Loop=App->StartOpenLoop(Form);
	if(!Loop) {
		delete Editor;
		delete Menu;
		delete Form;
		delete Font;
		delete App;
		memset((char *)this,0,sizeof(*this));
		return;
		
	}
	
	Editor->SetReadOnly(TRUE);
	Editor->SetOverwriteMode(TRUE);

	update_event=tmr_eventcreate(updateConsole,250,TRUE);

}

STDConsole::~STDConsole()
{
	lock=1;
	if(Editor) delete Editor;
	if(Menu) delete Menu;
	if(Form) delete Form;
	if(App) delete App;
	if(Font) delete Font;
	tmr_eventkill(update_event);

}

int STDConsole::put(const char *text)
{
	lock=1;
	Editor->InsertText(Editor->CursorPos,strlen(text),(char *)text);
	// CHECK FOR MAX. LENGTH EXCEEDED
	if(MaxSize>0) {
	Editor->LimitTextSize(MaxSize,TRIM_BEGIN);
	}
	App->Update();
	lock=2;
	return strlen(text);
}

void STDConsole::ConsoleBreak()
{
	lock=1;
	exit(0);
}


BOOL CatchKeyHandler(gEvent *e,gControl *This)
{
Console->LastKey=e->Data;
e->Data=FALSE;		// NO CONSOLE ECHO
return TRUE;
}

BOOL ViewMenuHandler(gEvent *e,gControl *This)
{
	if(Console->lock==-1) return TRUE;
	Console->lock=1;
	int EditMin,EditMax,Cursor;
	BOOL readonly;
	Console->Menu->ClearItems();
	Console->Menu->AddItemList((gSoftMenuItem *)ViewModeMenu,6);
	gApplication *App=This->GetApplication();
	
	Console->Editor->GetEditLimits(&EditMin,&EditMax);
	readonly=Console->Editor->IsReadOnly();
	Cursor=Console->Editor->CursorPos;
	Console->Editor->SetReadOnly(TRUE);
	Console->Editor->ClearEditLimits();
	
	
	while(App->RunOpenLoopOnce(Console->Loop));
	
	Console->Editor->SetReadOnly(readonly);
	Console->Editor->SetEditLimits(EditMin,EditMax);
	Console->Editor->MoveCursor(Cursor);
	
	if(App->RunResult==100) {	// EXIT MENU HANDLER
		App->RunLoopFlags&=~GAPP_EXIT;	// REMOVE THE EXIT REQUEST
		// BACK TO NORMAL MODE
		Console->Menu->ClearItems();
		Console->Menu->AddItemList((gSoftMenuItem *)NormalModeMenu,6);
		
	}

	Console->lock=0;
return TRUE;
}

BOOL ExitViewMenuHandler(gEvent *e,gControl *This)
{
	This->GetApplication()->Terminate(100);
return TRUE;
}


BOOL BreakHandler(gEvent *e,gControl *This)
{
	e->Data=TRUE;
	if(This->GetApplication()->QueryTerminate())
	{
		This->GetApplication()->Terminate(0);
	}
	return TRUE;
}

BOOL BreakQueryHandler(gEvent *e,gControl *This)
{
	if(Console->lock==-1) return FALSE;
	int s=gDoMessageBox(This->GetApplication(),"Console BREAK","BREAK requested,\nexit program?",S_YESNO);
	if(s==S_YES) return TRUE;
	return FALSE;
}

int STDConsole::getchar()
{
	lock=1;
	Editor->SetReadOnly(FALSE);
	Editor->SetEditLimits(Editor->CursorPos,Editor->CursorPos);
	Editor->AddHandler(GEVENT_GUI,-1,GEVGUI_BEFORECHANGE,&CatchKeyHandler);

	LastKey=0;

	do {
		if(!App->RunOpenLoopOnce(Loop)) break;
	} while (LastKey==0);
	
	if(LastKey==0) ConsoleBreak(); 	// MEANS RUNOPENLOOPONCE TERMINATED = CONSOLE BREAK


	Editor->RemoveHandler(GEVENT_GUI,-1,GEVGUI_BEFORECHANGE);
	Editor->SetReadOnly(TRUE);
	lock=0;
	return LastKey;
}

BOOL EnterKeyHandler(gEvent *e,gControl *This)
{
Console->LastKey=1;
e->Data=TRUE;		// ALLOW CONSOLE ECHO
return TRUE;
}



char * STDConsole::gets(char *str)
{
	lock=1;
	Editor->SetReadOnly(FALSE);
	Editor->SetOverwriteMode(FALSE);
	Editor->SetEditLimits(Editor->CursorPos,Editor->CursorPos);
	Editor->AddHandler(GEVENT_GUI,-1,GEVGUI_ENTER,&EnterKeyHandler);

	LastKey=0;
	do {
		if(!App->RunOpenLoopOnce(Loop)) break;
	} while (LastKey==0);
	
	if(LastKey==0)
		ConsoleBreak(); 	// MEANS RUNOPENLOOPONCE TERMINATED = CONSOLE BREAK

	Editor->RemoveHandler(GEVENT_GUI,-1,GEVGUI_ENTER);
	
	int start,end;
	
	
	Editor->GetEditLimits(&start,&end);
	Editor->ClearEditLimits();
	Editor->SetReadOnly(TRUE);
	Editor->SetOverwriteMode(TRUE);

	if(Editor->CursorPos<end) end=Editor->CursorPos;
	
	Editor->GetText(str,end-start,start,end);
	str[end-start-1]=0;	// DO NOT RETURN THE NEWLINE

	lock=0;
	return str;
}








void STDConsole::FlushAllAndClose()
{
	lock=1;
	App->RunOpenLoopOnce(Loop);
	
	App->EndOpenLoop(Loop);
	
	App->CloseAllAndExit();

	
}


void STDConsole::SetSizeLimit(int limit)
{
	if(limit<0) MaxSize=-1;
	else MaxSize=limit;
}

void STDConsole::clear()
{
	lock=1;
	Editor->RemoveText(0,Editor->Strlen());
	App->Update();
	lock=0;
}


void STDConsole::clreol()
{
	lock=1;
	int eol=Editor->Text.EndOfLine(Editor->CursorPos);
	
	Editor->RemoveText(Editor->CursorPos,eol);
	App->Update();
	lock=0;
}



void STDConsole::gotoxy(int x,int y)
{
	lock=1;
	int line=0,eol;
	if(y<0) return;
	if(x<0) return;
	while(y--) {
		line=Editor->Text.NextLine(line);
		if(line<0) break;
	}
	if(line<0) {
		Editor->MoveCursor(Editor->Text.EndOfText());
		App->Update();
		return;
	}
	
	eol=Editor->Text.EndOfLine(line);
	if(eol-line<x) {
		Editor->MoveCursor(eol);
		while(eol-line<x) {
		Editor->InsertChar(' ');
		++eol;
	}

	}
	else {
		Editor->MoveCursor(line+x);
	}
	App->Update();
	lock=0;
}

void STDConsole::wherexy(int *x,int *y)
{

	lock=1;
	int line=0,cy=0;
	if(y<0) return;
	if(x<0) return;
	while(line<Editor->CursorLine) {
		line=Editor->Text.NextLine(line);
		if(line<0) break;
		++cy;
	}
	if(y!=NULL) *y=cy;
	if(x!=NULL) *x=Editor->CursorPos-Editor->CursorLine;
	lock=0;
}

extern "C" void ShutDownConsole()
{
	if(Console) {
		Console->FlushAllAndClose();
		delete Console;
		ConsolePTR=NULL;
	}
}

extern "C" void InitConsole()
{
	if(Console==NULL) {
	ConsolePTR=new STDConsole();
	atexit(&ShutDownConsole);
	return;
	}
}


extern "C" void updateConsole(void)
{
	if(!Console) return;
	if(Console->lock==1) return;
	if(Console->lock==2) {
		Console->Editor->ClearClipRegion();
		Console->Editor->DrawAllLines();
		Console->Editor->InvalidateVScrollBar();
	}
	Console->lock=-1;
	Console->App->RunOpenLoopOnce(Console->Loop);
	Console->lock=0;
}



extern "C" char * getsConsole(char *str)
{
	if(!Console) {
		InitConsole();
		if(!Console) return 0;
	}
	return Console->gets(str);
}

extern "C" int getcharConsole()
{
	if(!Console) {
		InitConsole();
		if(!Console) return 0;
	}
	return Console->getchar();
}
extern "C" int putsConsole(const char *text)
{
	int count;
	if(!Console) {
		InitConsole();
		if(!Console) return 0;
	}
	count=Console->put(text);
	Console->put("\n");
	return count+1;
}

extern "C" int puts_nonlConsole(const char *text)
{
	if(!Console) {
		InitConsole();
		if(!Console) return 0;
	}
	return Console->put(text);
}


extern "C" int putcharConsole(char c)
{
	if(!Console) {
		InitConsole();
		if(!Console) return 0;
	}
	if(c==0) return 0;
	char str[2]={0,0};
	str[0]=c;
	return Console->put(str);
}



extern "C" void clearConsole()
{
	if(!Console) {
		InitConsole();
		if(!Console) return;
	}
	Console->clear();
}

extern "C" void clreolConsole()
{
	if(!Console) {
		InitConsole();
		if(!Console) return;
	}
	Console->clreol();
}


extern "C" void gotoxyConsole(int x,int y)
{
	if(!Console) {
		InitConsole();
		if(!Console) return;
	}
	Console->gotoxy(x,y);
}

extern "C" void wherexyConsole(int *x,int *y)
{
	if(!Console) {
		InitConsole();
		if(!Console) return;
	}
	Console->wherexy(x,y);
}
#undef Console
