//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_rarray.c 547 2008-02-06 01:01:02Z claudio $

#include <hpgcc3.h>
#include <master_globals.h>


SAT_OBJECT sat3_createrarray(SAT_OBJECT where,int rows,int cols)
{
int dims=( (rows && cols)? 2: (rows || cols)? 1:0);
int size;
SAT_OBJECT ptr;

	if(!dims) return 0;
	if(dims==1) { rows+=cols; cols=1; }		// SUPPORT FOR ONE-DIMENSIONAL VECTORS BY MAKING EITHER ROWS OR COLS =0
	
	size=20+5*dims+16*rows*cols;
    where=sat3_alloc(where,size);
	if(where==0) return 0;
	
	sat3_poke(where,SAT_DOARRY,5);
    sat3_poke(where+5,size-5,5);
	sat3_poke(where+10,SAT_DOREAL,5);
	ptr=where+15;
	sat3_poke(ptr,dims,5);
	sat3_poke(ptr+5,rows,5);
	ptr+=10;
	if(dims==2) { sat3_poke(ptr,cols,5); ptr+=5; }
	sat3_memset(ptr,0,8,where+size-ptr);

    return where;
}





extern ULONGLONG __sat3_dbl2real(double value);
extern ULONGLONG __sat3_xreal2real(XREAL *value);
extern ULONGLONG __sat3_decn2real(decNumber *value);

extern double __sat3_real2dbl(SAT_OBJECT realbody);
extern void __sat3_real2xreal(SAT_OBJECT realbody,XREAL *result);
extern void __sat3_real2decn(SAT_OBJECT realbody,decNumber *result);

SAT_OBJECT __sat3_arrayget(SAT_OBJECT array,int row,int col)
{
	if(!array) return 0;
int len,rows,cols,dims,type=0;
SAT_OBJECT end;
	if(sat3_peek(array,5)!=SAT_DOARRY) return 0;		// RETURN IF NOT AN ARRAY
	type=sat3_typesize(sat3_peek(array+10,5));			// GET SIZE OF THE ATOMIC TYPE OF THE ARRAY
	if(!type) return 0;
	type-=5;
	len=sat3_peek(array+5,5);
	dims=sat3_peek(array+15,5);
	if(dims<1 || dims>2) return 0;					// ONLY 1 OR 2 DIMS SUPPORTED
	end=array+len+5;
	array+=20;
	rows=sat3_peek(array,5);
	if(dims>=2) { 
		array+=5; cols=sat3_peek(array,5);
		if(row>=rows || row<0) return 0;
		if(col>=cols || col<0) return 0;
	}
	else {
		if(row>=rows || row<0) return 0;
		if(col>=rows || col<0) return 0;
		if(row*col) return 0;			// ONE OF THEM HAS TO BE ZERO
		cols=1;
	}
	
	array+=5;
	array+=(col+row*cols)*type;
	if(array>end-type) return 0;						// IN CASE NUMBER OF ROWS AND COLUMNS IS INVALID
	return array;
}

BOOL sat3_israrray(SAT_OBJECT array)
{
	if(!array) return 0;
	if(sat3_peek(array,5)!=SAT_DOARRY) return 0;		// RETURN IF NOT AN ARRAY
	if(sat3_peek(array+10,5)==SAT_DOREAL) return 1;
	return 0;
}


void sat3_rarrayput_dbl(SAT_OBJECT array,int row,int col,double value)
{
ULONGLONG real;

	if(!sat3_israrray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_dbl2real(value);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);

}



void sat3_rarrayput_xreal(SAT_OBJECT array,int row,int col,XREAL *value)
{
ULONGLONG real;

	if(!sat3_israrray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_xreal2real(value);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);

}



void sat3_rarrayput_int(SAT_OBJECT array,int row,int col,int value)
{
XREAL x;

xr_int2xr(value,&x);

sat3_rarrayput_xreal(array,row,col,&x);


}

void sat3_rarrayput_llong(SAT_OBJECT array,int row,int col,LONGLONG value)
{
XREAL x;

xr_llong2xr(value,&x);

sat3_rarrayput_xreal(array,row,col,&x);


}

void sat3_rarrayput_decn(SAT_OBJECT array,int row,int col,decNumber *value)
{
ULONGLONG real;

	if(!sat3_israrray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_decn2real(value);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);

}

#define MAX_IEEE754_EXPO 307





double sat3_rarrayget_dbl(SAT_OBJECT array,int row,int col)
{

	if(!sat3_israrray(array)) return 0.0;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return 0.0;
	
	return __sat3_real2dbl(array);

}


BOOL sat3_rarrayget_xreal(SAT_OBJECT array,int row,int col,XREAL *result)
{

	if(!sat3_israrray(array)) {
	memset((void *)result,0,sizeof(XREAL));
	return FALSE;
	}
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) {
	memset((void *)result,0,sizeof(XREAL));
	return FALSE;
	}
	
	
	__sat3_real2xreal(array,result);
	return TRUE;

}

int sat3_rarrayget_int(SAT_OBJECT array,int row,int col)
{
XREAL x;
	if(sat3_rarrayget_xreal(array,row,col,&x)) return xr_xr2int(&x);
	return 0;
}

LONGLONG sat3_rarrayget_llong(SAT_OBJECT array,int row,int col)
{
XREAL x;
	if(sat3_rarrayget_xreal(array,row,col,&x)) return xr_xr2llong(&x);
	return 0;
}


BOOL sat3_rarrayget_decn(SAT_OBJECT array,int row,int col,decNumber *result)
{

	if(!sat3_israrray(array)) {
	memset((void *)result,0,sizeof(decNumber));
	return FALSE;
	}
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) {
	memset((void *)result,0,sizeof(decNumber));
	return FALSE;
	}
	
	
	__sat3_real2decn(array,result);
	return TRUE;

}

BOOL sat3_rarraysize(SAT_OBJECT array,int *row,int *col)
{
	if(!sat3_israrray(array)) return 0;
	int rows,cols,dims;

	dims=sat3_peek(array+15,5);
	if(dims<1 || dims>2) return 0;					// ONLY 1 OR 2 DIMS SUPPORTED
	array+=20;
	rows=sat3_peek(array,5);
	if(dims>=2) { array+=5; cols=sat3_peek(array,5); }
	else cols=0;
	*row=rows;
	*col=cols;
	
	return TRUE;
}
