//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_zint.c 557 2008-02-21 00:58:51Z claudio $

#include <hpgcc3.h>
#include <master_globals.h>








SAT_OBJECT sat3_createzint_xreal(SAT_OBJECT where,XREAL *value)
{
XREAL xr;
int zdigits;
int ndigits;
int sign;
int f;
char tmpbuff[22];


	sign=value->Sign;

	if(value->Mant==0LL) {
	zdigits=0;
	sign=0;
	ndigits=0;
	}
	else {
	xr_base10(value,&xr);		// CONVERT EXPONENT TO BASE-10 FORMAT
	
	// HERE WE HAVE SIGN, INTEGER MANTISSA W/16 DECIMAL DIGITS (left-justified) + EXPONENT IN BASE-10
	memset(tmpbuff,48,22);

	ndigits=xr_getdigits10(&xr,tmpbuff);
	
	zdigits=ndigits+xr.Exp;
	if(zdigits<0) { zdigits=0; sign=0; ndigits=0; }

}

    where=sat3_alloc(where,10+zdigits+1);
	if(where==0) return 0;
    
 	sat3_poke(where,SAT_DOINT,5);
	sat3_poke(where+5,zdigits+1+5,5);
	f=where+10;
	while(zdigits>0) {
	if(zdigits>ndigits) {
		sat3_poke(f,0,1);
	}
	else sat3_poke(f,tmpbuff[zdigits-1]-'0',1);
	
	++f;
	--zdigits;
	}
    sat3_poke(f,sign? 9:0,1);

    return where;
}


#define MAX_IEEE754_EXPO 307


// LOW-LEVEL CONVERSION
// IF zdigits<0 THEN zintbody POINTS TO THE BODY OF A ZINT OBJECT (TO SIZE MEMBER)
// IF zdigits>=0 THEN zintbody POINTS TO A zdigits LONG UNSIGNED BCD STRING IN MEMORY
// zexp IS EXPONENT IN BASE 10, THE zint WILL BE TREATED AS (ZINT*10^ZEXP)


void __sat3_zint2xreal(SAT_OBJECT zintbody,int zdigits,int zexp,XREAL *result)
{
	int sign=0;

	if(zdigits<0) {
	// PROCESS AS A ZINT OBJECT
	zdigits=sat3_peek(zintbody,5)-6;
	zintbody+=5;
	sign=(sat3_peek(zintbody+zdigits, 1)== 9) ? 1 : 0;	
	}
	
    // CONVERT A BCD STRING OF DIGITS
	
	if(!zdigits) {
	result->Mant=0;
	result->Exp=-63;
	result->Sign=0;
	return;
	}
	
	if(zdigits>18) {
	// RESURSIVELY PROCESS LARGE NUMBERS
	XREAL  hi,lo;
	int lodigits;
	__sat3_zint2xreal(zintbody+zdigits-18,18,zexp+zdigits-18,&hi);
	if(zdigits>26)	lodigits=8;	// 26 DIGITS TO ENSURE CORRECT ROUNDING OF LAST BIT
	else lodigits=zdigits-18;
	__sat3_zint2xreal(zintbody+zdigits-lodigits-18,lodigits,zexp+zdigits-18-lodigits,&lo);
	xr_add(&hi,&lo,result);
	result->Sign=sign;
	return;
	}
	
	result->Mant = (ULONGLONG)sat3_bcd2llong(zintbody,zdigits);
	result->Exp=zexp;
	result->Sign=sign;
	xr_base2(result,result);
	return;
}




BOOL sat3_convertzint_xreal(SAT_OBJECT zint,XREAL *result)
{
if(!zint) return 0;
if(sat3_isptr(zint)) zint=sat3_peek(zint,5);

if(sat3_peek(zint,5)!=SAT_DOINT) return 0;

zint+=5;
    
	__sat3_zint2xreal(zint,-1,0,result);

	 return TRUE;
	 
}


SAT_OBJECT sat3_createzint_decn(SAT_OBJECT where,decNumber *value)
{
int size,udigits,digitcount,skipdigits;
SAT_OBJECT addr;

// WARNING: ASSUMING DECDPUN<=16, WILL FAIL IF NOT
	decNumberUnit *unitptr,ubcd,unit;

	int f;
	size=value->digits+value->exponent;
	if(value->exponent<0) skipdigits=-value->exponent;
	else skipdigits=0;
	
	
	if(size==1 && value->lsu==0) {
	// NUMBER IS ZERO, THEREFORE CONVERT TO ZINT ZERO = 0 DIGITS
	size=0;
	}
    where=sat3_alloc(where,11+size);
	if(where==0) return 0;
    
	sat3_poke(where,SAT_DOINT,5);
    sat3_poke(where+5,6+size,5);
	sat3_poke(where+5+size+5,(value->bits&DECNEG)? 9:0,1);
	if(!size) return where;
	
	decNumberUnit pow10[DECDPUN+1];
	// 10 POWERS TO CONVERT UNIT
    pow10[0]=1;
	for(f=1;f<=DECDPUN;++f) pow10[f]=pow10[f-1]*10;
	
	unitptr=value->lsu;

	addr=where+10;
	
	digitcount=0;
	
	while(size) {
	if(digitcount>=value->digits) {
	// ADD ZEROES AT THE END OF NUMBER
	ubcd=0;
	udigits=(size>DECDPUN)? DECDPUN:size;
	}
	else {

	if(skipdigits>=DECDPUN) { ++unitptr; skipdigits-=DECDPUN; digitcount+=DECDPUN; continue; }
	udigits=(value->digits-digitcount>DECDPUN)? DECDPUN : (value->digits-digitcount);
	unit=*unitptr++;
	
	ubcd=0;
	for(f=DECDPUN;f>=0;f--) {
	ubcd<<=4;
	while(unit>=pow10[f]) { ubcd++; unit-=pow10[f]; }
	}
	if(skipdigits) { ubcd>>=skipdigits<<2; udigits-=skipdigits; digitcount+=skipdigits; skipdigits=0; }
	
	
	}
	sat3_poke(addr,ubcd,udigits);
	size-=udigits;
	addr+=udigits;
	digitcount+=udigits;
	}
	
    return where;
}

// THIS ROUTINE ALLOCATES ITS OWN DECNUMBER, IT BREAKS CONSISTENCY!!!
// NO OTHER WORKAROUND POSSIBLE DUE TO ARBITRARY LENGTH IN ZINTS

BOOL sat3_convertzint_decn(SAT_OBJECT zint,decNumber **result)
{
	int pow10;
	if(!zint) return 0;
	if(sat3_isptr(zint)) zint=sat3_peek(zint,5);

decNumberUnit *ptr;
if(sat3_peek(zint,5)!=SAT_DOINT) return 0;

zint+=5;

int digits=sat3_peek(zint,5)-6,udigits;

if(digits==0) {
	digits=1;
	--zint;		// HACK: THE NIBBLE RIGHT BEFORE THE SIGN IS ALWAYS ZERO WHEN digits IS ZERO
}

	zint+=5;
    int numsize=(digits+DECDPUN-1)/DECDPUN*sizeof(decNumberUnit) + 2*sizeof(int32_t) + sizeof(uint8_t);
    if(numsize<sizeof(decNumber)) numsize=sizeof(decNumber);
    *result = (decNumber *)malloc(numsize);
	
	if(!*result)  return 0;
	
	(*result)->digits = digits;
	(*result)->bits=(sat3_peek(zint+digits,1)!=0)? DECNEG:0;
	(*result)->exponent=0;
	ptr=(*result)->lsu;

	while(digits)  {
	udigits=(digits>DECDPUN)? DECDPUN:digits;
	*ptr=0;
	pow10=1;
	while(udigits) {
	*ptr+=sat3_peek(zint,1)*pow10;
	udigits--;
	digits--;
	zint++;
	pow10*=10;
	}
	++ptr;
	
	}

	return TRUE;
    
}


SAT_OBJECT sat3_createzint_dbl(SAT_OBJECT where,double value)
{
XREAL xr;

xr_double2xr(value,&xr);

return sat3_createzint_xreal(where,&xr);
}


double sat3_convertzint_dbl(SAT_OBJECT zint)
{
XREAL xr;

if(sat3_convertzint_xreal(zint,&xr)) return xr_xr2double(&xr);

return 0.0;
}


SAT_OBJECT sat3_createzint_int(SAT_OBJECT where,int value)
{
XREAL xr;

xr_int2xr(value,&xr);

return sat3_createzint_xreal(where,&xr);
}


int sat3_convertzint_int(SAT_OBJECT real)
{
XREAL xr;

if(sat3_convertzint_xreal(real,&xr)) return xr_xr2int(&xr);

return 0;
}



SAT_OBJECT sat3_createzint_llong(SAT_OBJECT where,LONGLONG value)
{
XREAL xr;

xr_llong2xr(value,&xr);

return sat3_createzint_xreal(where,&xr);
}


LONGLONG sat3_convertzint_llong(SAT_OBJECT real)
{
XREAL xr;

if(sat3_convertzint_xreal(real,&xr)) return xr_xr2llong(&xr);

return 0LL;
}

